#
# Copyright © 2019 Paulo Henrique Silva <ph.silva@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#
# Check assert success
#

assert 42 == 42
assert True
assert 1
assert [42]
assert "true"
assert "false"  # string s is True if len(s) > 0
